#include <stdio.h>
#include <unistd.h>
#include <syslog.h>
#include <error.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <linux/ioctl.h>

#include "ksaf_devctl.h"


#define OP_TYPE_ALLOW		1
#define OP_TYPE_DISALLOW	2
#define OP_TYPE_ALLOW_ONCE	3

#define KSAF_CHAR_MID_LEN 128
#define KSAF_IOCTL_TYPE 0xaf
#define KSAF_POLICY_QUERY _IOWR(KSAF_IOCTL_TYPE, 1, struct ksaf_device_context)

#define KSAF_STATUS_FILE "/sys/kernel/security/ksaf/status"

struct ksaf_device_context {
	uint32_t result;
	uint32_t uid;
	uint64_t sid;
	uint64_t oid;
	uint32_t perm_class_id;
	uint32_t perm_id;
	uint32_t pid;
	uint32_t vid;
	uint32_t if_protocol;
	char serial[KSAF_CHAR_MID_LEN];
} __attribute__((aligned(4)));


enum kylin_printer_type {
	KYLIN_PRINTER = 1,
	KYLIN_PRINTER_LOCAL,
	KYLIN_PRINTER_NET,
};

const char *printer_str[] = {
	"none",
	"printer",
	"local printer(usb/serial/lp)",
	"net printer",
};

// typedef struct {
//     unsigned int (*bluetooth_dev_control)(unsigned int dev_type, const char *dev_addr);
// } PluginInterface;

// static unsigned int test_dev_control(unsigned int dev_type, const char *dev_addr)
// {
// 	openlog("ksaf-devctl",LOG_PID,LOG_LOCAL5);
// 	syslog(LOG_WARNING, "%s: dev_type=%d, pid=%d", __func__, dev_type, getpid());

// 	return 0;
// }

// PluginInterface plugin_interface = {
//     .bluetooth_dev_control = test_dev_control,
// };

void ksaf_devctl_log(int level, const char *format, ...)
{
	va_list vlist;

	va_start(vlist, format);

	openlog("ksaf-devctl", LOG_PID, LOG_LOCAL5);
	vsyslog(level, format, vlist);
	closelog();

	va_end(vlist);
}

int ksaf_get_status(void)
{
	int status = 0;
	FILE *fp = NULL;

	if (access(KSAF_STATUS_FILE, F_OK) != 0) {
		ksaf_devctl_log(LOG_WARNING, "%s: ksaf is not support!", __func__);
		return -ENOENT;
	}

	fp = fopen(KSAF_STATUS_FILE, "r");
	if (fp != NULL) {
		status = fgetc(fp) - '0';
		fclose(fp);
	}
	return status;
}

static int ksaf_devctl_ioctl_query(struct ksaf_device_context *recv)
{
	int fd = 0;
	int ret= 0;

	fd = open("/dev/ksaf", O_RDWR);
	if (fd < 0) {
		ksaf_devctl_log(LOG_ERR, "open /dev/ksaf failed, errno=%d", errno);
		return 0;
	}

	ret = ioctl(fd, KSAF_POLICY_QUERY, recv);
	if (ret < 0) {
		ksaf_devctl_log(LOG_ERR, "%s: ioctl error, errno=%d", errno);
		return 0;
	}
	ret = recv->result;
	close(fd);
	return ret;
}

static int ksaf_devctl_query_perm(uint64_t sid,
		uint32_t perm_class_id, uint32_t perm_id)
{
	struct ksaf_device_context info = {0};
	int ret = 0;

	memset(&info, 0x00, sizeof(info));
	info.uid = getuid();
	info.sid = sid;
	info.perm_class_id = 20 + perm_class_id;
	info.perm_id = perm_id;

	ret = ksaf_devctl_ioctl_query(&info);
	if (ret == OP_TYPE_DISALLOW)
		return -EACCES;

	return 0;
}

static int switch_printer_type(int type)
{
	int printer_type = 0;

	switch (type) {
	case KYLIN_PRINTER:
		printer_type = KSAF_PRINTER;
		break;
	case KYLIN_PRINTER_LOCAL:
		printer_type = KSAF_PRINTER_LOCAL;
		break;
	case KYLIN_PRINTER_NET:
		printer_type = KSAF_PRINTER_NET;
		break;
	default:
		break;
	}
	return printer_type;
}

static int get_comm_by_proc(pid_t pid)
{
	FILE *fp = NULL;
	char path[30] = {0};
	char buffer[256] = {0};

	snprintf(path, sizeof(path), "/proc/%d/comm", pid);
	fp = fopen(path, "r");
	if (fp == NULL) {
		ksaf_devctl_log(LOG_ERR, "%s: open %s failed", __func__, path);
		return 0;
	}
	if (fgets(buffer, sizeof(buffer), fp) != NULL && buffer[0] != '\0')  {
		ksaf_devctl_log(LOG_WARNING, "%s: pid=%d, comm=%s", __func__, pid, buffer);
		buffer[sizeof(buffer) - 1] = '\0';
		if (!strncmp(buffer, "kylin-video", 11) ||
				!strncmp(buffer, "kylin-record", 12)) {
			ksaf_devctl_log(LOG_WARNING, "%s: comm=%s is blocked!\n",
					__func__, buffer);
			return -EACCES;
		}
	}
	return 0;
}

static uint64_t get_sid_by_pid(pid_t pid)
{
	FILE *fp = NULL;
	char path[30] = {0};
	char buffer[256] = {0};
	uint64_t sid = 0;
	char a_sid[20] = {0};
	char t_id[20] = {0};
	char exec_sid[20] = {0};

	snprintf(path, sizeof(path), "/proc/%d/attr/ksaf/current", pid);
	fp = fopen(path, "r");
	if (fp == NULL) {
		printf("%s: open path failed", __func__);
		return -1;
	}

	if (fgets(buffer, sizeof(buffer), fp) != NULL && buffer[0] != '\n')  {
		sscanf(buffer, "%[^:]:%[^:]:%[^:]:", a_sid, t_id,exec_sid);
		sid = strtoul(exec_sid, NULL, 16);
	}

	return sid;
}

int ksaf_get_printer_perm_by_type(int type)
{
	int perm = 0;
	uint64_t sid = 0;
	int printer_type = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	printer_type = switch_printer_type(type);
	if (printer_type <= 0) {
		ksaf_devctl_log(LOG_DEBUG, "%s: printer_type is wrong, type=%d",
			__func__, type);
		return 0;
	}

	perm = ksaf_devctl_query_perm(0, printer_type, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: printer=%s perm=%d",
			__func__, printer_str[type], perm);
	return perm;
}

int ksaf_get_cups_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KSAF_PRINTER, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: pid=%d perm=%d", __func__, pid, perm);
	return perm;
}

int ksaf_get_audio_jack_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KSAF_AUDIO_JACK, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: pid=%d perm=%d", __func__, pid, perm);
	return perm;
}

int ksaf_get_audio_playback_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KASF_AUDIO_SPEAKER, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: sid=%16lx pid=%d perm=%d", __func__, sid, pid, perm);
	return perm;
}

int ksaf_get_audio_record_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KSAF_AUDIO_MIC, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: pid=%d perm=%d", __func__, pid, perm);
	return perm;
}

int ksaf_get_sane_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KSAF_SCANNER, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: sid=%16lx pid=%d perm=%d", __func__, sid, pid, perm);
	return perm;
}

int ksaf_get_wince_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KSAF_WINCE, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: sid=%16lx pid=%d perm=%d", __func__, sid, pid, perm);
	return perm;
}

int ksaf_get_bluetooth_perm_by_pid(pid_t pid)
{
	uint64_t sid = 0;
	int perm = 0;

	if (ksaf_get_status() <= 0)
		return 0;

	sid = get_sid_by_pid(pid);
	perm = ksaf_devctl_query_perm(sid, KSAF_BLUETOOTH, KSAF_DEVICE_USE);
	ksaf_devctl_log(LOG_DEBUG, "%s: sid=%16lx pid=%d perm=%d", __func__, sid, pid, perm);
	return perm;
}
